import random
import numpy as np
import networkx as nx


def preferences_ut_to_preferences_rk(preferences_ut):
    """Convert utilities to rankings.

    Parameters
    ----------
    preferences_ut : list of list of numbers (or ndarray)
        ``preferences_ut[v, c]`` is the utility of candidate ``c`` as seen by voter ``v``.

    Returns
    -------
    preferences_rk : ndarray
        ``preferences_rk[v, k]`` is the candidate at rank ``k`` for voter ``v``. If
        ``preferences_ut[v,c] == preferences_ut[v,d]``, then it is drawn at random whether ``v`` prefers
        ``c`` to ``d``  or ``d`` to ``c``.

    Examples
    --------
        >>> preferences_ut_to_preferences_rk(preferences_ut=[[5, 1, 2], [4, 10, 1]])
        array([[0, 2, 1],
               [1, 0, 2]])
    """
    preferences_ut = np.array(preferences_ut)
    n_v, n_c = preferences_ut.shape
    tiebreaker = np.random.rand(n_v, n_c)
    return np.array(np.lexsort((tiebreaker, - preferences_ut), 1), dtype=int)


def preferences_rk_to_preferences_borda_rk(preferences_rk):
    """Convert rankings to Borda scores.

    Parameters
    ----------
    preferences_rk : list of list of integers (or ndarray)
        ``preferences_rk[v, k]`` is the candidate at rank ``k`` for voter ``v``.

    Returns
    -------
    preferences_borda_rk : ndarray
        ``preferences_borda_rk[v, c]`` is the Borda score (between 0 and ``C - 1``) of candidate ``c`` for voter ``v``.

    Examples
    --------
        >>> preferences_rk_to_preferences_borda_rk(preferences_rk=[[0, 2, 1], [1, 0, 2]])
        array([[2, 0, 1],
               [1, 2, 0]])
    """
    preferences_rk = np.array(preferences_rk)
    _, n_c = preferences_rk.shape
    return - np.array(np.argsort(preferences_rk, 1), dtype=int) + n_c - 1


def preferences_ut_to_preferences_borda_ut(preferences_ut):
    """Convert utilities to Borda scores (with possible equalities).

    Parameters
    ----------
    preferences_ut : list of list of numbers (or ndarray)
        ``preferences_ut[v, c]`` is the utility of candidate ``c`` as seen by voter ``v``.

    Returns
    -------
    preferences_borda_ut : ndarray
        ``preferences_borda_ut[v, c]`` gains 1 point for each ``d`` such that ``v`` prefers ``c`` to ``d``,
        and 0.5 point for each ``d`` such that ``v`` is indifferent between ``c`` and ``d``.

    Examples
    --------
        >>> preferences_ut_to_preferences_borda_ut(preferences_ut=[[5, 1, 2], [4, 10, 1]])
        array([[2., 0., 1.],
               [1., 2., 0.]])
    """
    preferences_ut = np.array(preferences_ut)
    n_v, n_c = preferences_ut.shape
    preference_borda_ut = np.zeros((n_v, n_c))
    for c in range(n_c):
        preference_borda_ut[:, c] = np.sum(
            0.5 * (preferences_ut[:, c][:, np.newaxis] >= preferences_ut)
            + 0.5 * (preferences_ut[:, c][:, np.newaxis] > preferences_ut),
            axis=1
        ) - 0.5
    return preference_borda_ut


def preferences_ut_to_matrix_duels_ut(preferences_ut):
    """Compute the matrix of duels.

    Parameters
    ----------
    preferences_ut : list of list of numbers (or ndarray)
        ``preferences_ut[v, c]`` is the utility of candidate ``c`` as seen by voter ``v``.

    Returns
    -------
    matrix_duels_ut : ndarray
        ``matrix_duels_ut[c, d]`` is the number of voters who strictly prefer candidate ``c`` to candidate ``d``.
        By convention, diagonal coefficients are set to 0.

    Examples
    --------
        >>> preferences_ut_to_matrix_duels_ut(preferences_ut=[[5, 1, 2], [4, 10, 1]])
        array([[0, 1, 2],
               [1, 0, 1],
               [0, 1, 0]])
    """
    preferences_ut = np.array(preferences_ut)
    n_v, n_c = preferences_ut.shape
    matrix_duels = np.zeros((n_c, n_c), dtype=np.int)
    for c in range(n_c):
        for d in range(c + 1, n_c):
            matrix_duels[c, d] = np.sum(preferences_ut[:, c] > preferences_ut[:, d])
            matrix_duels[d, c] = np.sum(preferences_ut[:, d] > preferences_ut[:, c])
    return matrix_duels


def matrix_victories_to_smith_set(matrix_victories):
    """
    Smith set associated to a matrix of victories.

    Parameters
    ----------
    matrix_victories : ndarray.
        ``matrix_victories[c, d]`` is 1 vs 0 in case of victory, 0.5 vs 0.5 in case of tie. Diagonal coefficients
        can use any conventional value, but they must all have the same value.

    Returns
    -------
    smith_set : list
        List of candidates in the Smith set, sorted by increasing order.

    Examples
    --------
        >>> matrix_victories = np.array([
        ...     [0, 1,  1, 1, 1,  1, 0],
        ...     [0, 0,  0, 0, 1,  0, 0],
        ...     [0, 1,  0, 0, 1, .5, 1],
        ...     [0, 1,  1, 0, 1,  1, 1],
        ...     [0, 0,  0, 0, 0,  0, 0],
        ...     [0, 1, .5, 0, 1,  0, 0],
        ...     [1, 1,  0, 0, 1,  1, 0]
        ... ])
        >>> matrix_victories_to_smith_set(matrix_victories)
        [0, 2, 3, 5, 6]
    """
    n_c = matrix_victories.shape[0]
    copeland_scores = matrix_victories.sum(axis=1)
    copeland_decreasing = sorted(copeland_scores, reverse=True)
    candidates_by_copeland_best_worst = sorted(range(n_c), key=copeland_scores.__getitem__, reverse=True)
    matrix_victories_sorted = (
        matrix_victories[candidates_by_copeland_best_worst, :][:, candidates_by_copeland_best_worst])
    i = 0
    while True:
        while i < n_c - 1 and copeland_decreasing[i] == copeland_decreasing[i + 1]:
            i += 1
        if np.all(matrix_victories_sorted[0:i + 1, i + 1:] == 1):
            return sorted(candidates_by_copeland_best_worst[0:i + 1])
        i += 1


def is_resistant_condorcet(w, preferences_ut):
    """Test for resistant Condorcet winner.

    A Condorcet winner ``w`` is *resistant* iff in any Condorcet voting system, the profile is not manipulable (cf.
    Durand et al. 2014). This is equivalent to say that for any pair ``(c, d)`` of other candidates, there is a strict
    majority of voters who simultaneously:

    * Do not prefer ``c`` to ``w``,
    * And prefer ``w`` to ``d``.

    Parameters
    ----------
    w : int (candidate) or NaN
    preferences_ut : list of list of numbers (or ndarray)
        ``preferences_ut[v, c]`` is the utility of candidate ``c`` as seen by voter ``v``.

    Returns
    -------
    bool
        If w is a resistant Condorcet winner, then True. Otherwise (or if w is NaN), then False.

    Examples
    --------
        >>> is_resistant_condorcet(w=0, preferences_ut=[[5, 1, 2], [4, 10, 1]])
        False
    """
    if np.isnan(w):
        return False
    preferences_ut = np.array(preferences_ut)
    n_v, n_c = preferences_ut.shape
    for c in range(n_c):
        if c == w:
            continue
        v_does_not_prefer_c_to_w = (preferences_ut[:, w] >= preferences_ut[:, c])
        for d in range(n_c):
            if d == w:
                continue
            v_prefers_w_to_d = (preferences_ut[:, w] > preferences_ut[:, d])
            if np.sum(np.logical_and(v_does_not_prefer_c_to_w, v_prefers_w_to_d)) <= n_v / 2:
                return False
    return True


def compute_next_subset_with_w(prev_subset, n_c, n_c_r, w):
    """Compute the next subset containing w, by lexicographic order.

    This function is internally used by :class:`Rule` to compute Independence of Irrelevant Alternatives (IIA).

    Parameters
    ----------
    prev_subset : list of integers (or 1d ndarray)
        ``prev_subset(k)`` is the ``k``-th candidate of the subset. Candidates must be sorted by ascending order.
        Candidate ``w`` must belong to ``prev_subset``.
    n_c : int
        Total number of candidates.
    n_c_r : int
        Number of candidates for the subset.
    w : int
        A candidate whose presence is required in the subset.

    Returns
    -------
    next_subset : list of integers
        ``next_subset(k)`` is the ``k``-th candidate of the subset. Candidates are sorted by ascending order.
        Candidate ``w`` belongs to next_subset.

    Examples
    --------
        >>> compute_next_subset_with_w(prev_subset=[0, 2, 7, 8, 9], n_c=10, n_c_r=5, w=0)
        [0, 3, 4, 5, 6]
        >>> compute_next_subset_with_w(prev_subset=[0, 2, 7, 8, 9], n_c=10, n_c_r=5, w=8)
        [0, 3, 4, 5, 8]

    If we have already the last subset, then the result is None:

        >>> print(compute_next_subset_with_w(prev_subset=[0, 6, 7, 8, 9], n_c=10, n_c_r=5, w=0))
        None
    """
    # TODO: this could be rewritten with an iterator.
    max_allowed_value = n_c
    for index_pivot in range(n_c_r - 1, -1, -1):
        max_allowed_value -= 1
        if max_allowed_value == w:
            max_allowed_value -= 1
        if prev_subset[index_pivot] == w:
            max_allowed_value += 1
        elif prev_subset[index_pivot] < max_allowed_value:
            break  # Found the pivot
    else:
        return None
    next_subset = prev_subset
    new_member = prev_subset[index_pivot] + 1
    for i in range(index_pivot, n_c_r - 1):
        next_subset[i] = new_member
        new_member += 1
    next_subset[n_c_r - 1] = max(w, new_member)
    return next_subset


def compute_next_permutation(prev_permutation, n_c):
    """Compute next permutation by lexicographic order.

    Parameters
    ----------
    prev_permutation : list of integers (or ndarray)
        A list of distinct numbers.
    n_c : int
        Number of elements in prev_permutation

    Returns
    -------
    next_permutation : ndarray
        If ``prev_permutation`` was the last permutation in lexicographic order (i.e. all numbers sorted in descending
        order), then ``next_permutation = None``.

    Examples
    --------
        >>> compute_next_permutation([0, 2, 1, 4, 3], 5)
        array([0, 2, 3, 1, 4])
        >>> import numpy as np
        >>> compute_next_permutation(np.array([0, 2, 1, 4, 3]), 5)
        array([0, 2, 3, 1, 4])
    """
    # TODO: this could be rewritten with an iterator.
    prev_permutation = list(prev_permutation)
    for i in range(n_c - 2, -1, -1):
        if prev_permutation[i] < prev_permutation[i + 1]:
            index_pivot = i
            break
    else:
        return None
    index_replacement = None
    for i in range(n_c - 1, index_pivot, -1):
        if prev_permutation[i] > prev_permutation[index_pivot]:
            index_replacement = i
            break
    return np.array(
        prev_permutation[0:index_pivot] + [prev_permutation[index_replacement]]
        + prev_permutation[n_c - 1:index_replacement:-1]
        + [prev_permutation[index_pivot]] + prev_permutation[index_replacement-1:index_pivot:-1],
        dtype=int
    )


def compute_next_borda_clever(prev_permutation, prev_favorite, n_c):
    """Compute next vector of Borda scores in 'clever' order.

    Parameters
    ----------
    prev_permutation : list of integers (or 1d ndarray)
        Must be exactly all integers from 0 to C - 1.
    prev_favorite : int
        Index of the preferred candidate. I.e., ``prev_permutation[prev_favorite]`` must be equal to ``C - 1``.
    n_c : int
        Number of elements in ``prev_permutation``.

    Returns
    -------
    (``next_permutation``, ``next_favorite``) : tuple
        ``next_permutation`` is a 1d array of all integers from 0 to ``C - 1``, . Next vector of Borda scores in the
        'clever' order. If ``prev_permutation`` was the last permutation in the 'clever' order, then
        ``next_permutation = None``. ``next_favorite`` is an Integer (or None); it is the new preferred candidate.

    Notes
    -----
    A vector of Borda scores is seen as two elements:

    1. A permuted list of Borda scores from 0 to ``C - 2``,
    2. The insertion of Borda score ``C - 1`` in this list.

    The 'clever' order sorts first by lexicographic order on (1), then by the position (2) (from last position to
    first position).

    To find the next permutation, if Borda score ``C - 1`` can be moved one step to the left, we do it. Otherwise,
    we take the next permutation of ``{0, ..., C - 2}`` in lexicographic order, and put Borda score ``C - 1`` at the
    end.

    In the 'clever' order, the first vector is ``[0, ..., C - 1]`` and the last one is ``[C - 1, ..., 0]``.

    When looking for manipulations (IM or UM principally), here is the advantage of using the 'clever' order instead of
    lexicographic order: in only the ``C`` first  vectors, we try all candidates as top-ranked choice. In many voting
    systems, this accelerates finding the manipulation.

    Examples
    --------
        >>> compute_next_borda_clever(prev_permutation=[0, 1, 4, 2, 3], prev_favorite=2, n_c=5)
        (array([0, 4, 1, 2, 3]), 1)
        >>> compute_next_borda_clever(prev_permutation=[4, 0, 1, 2, 3], prev_favorite=0, n_c=5)
        (array([0, 1, 3, 2, 4]), 4)

    At the end of the iteration, we return (None, None):

        >>> compute_next_borda_clever(prev_permutation=[4, 3, 2, 1, 0], prev_favorite=0, n_c=5)
        (None, None)
    """
    # TODO: this could be rewritten with an iterator.
    if prev_favorite > 0:
        next_favorite = prev_favorite - 1
        next_permutation = np.copy(prev_permutation)
        next_permutation[[prev_favorite, next_favorite]] = next_permutation[[next_favorite, prev_favorite]]
    else:
        try:
            next_permutation = np.concatenate((compute_next_permutation(prev_permutation[1:n_c], n_c - 1), [n_c - 1]))
            next_favorite = n_c - 1
        except ValueError:
            next_permutation = None
            next_favorite = None
    return next_permutation, next_favorite


def strong_connected_components(a):
    """Strong connected components of a digraph, sorted by topological order.

    Parameters
    ----------
    a : a list of lists or a numpy array
        The adjacency matrix: ``a[i, j] = True`` means that ``i`` has an edge toward ``j``.

    Returns
    -------
    list of sets
        For example, the first set is the top component of the digraph; in the case of a victory matrix, it is the
        top set (smallest subset of candidates who win against all the others).

    Examples
    --------
        >>> strong_connected_components(a=[
        ...     [False,  True, False,  True,  True],
        ...     [False, False,  True,  True,  True],
        ...     [ True, False, False,  True,  True],
        ...     [False, False, False, False,  True],
        ...     [False, False, False, False, False]
        ... ]) == [{0, 1, 2}, {3}, {4}]
        True
    """
    condensed = nx.condensation(nx.DiGraph(np.array(a)))
    return [condensed.nodes[component]['members'] for component in nx.topological_sort(condensed)]


def initialize_random_seeds(n=0):
    """Initialize the random seeds.

    Parameters
    ----------
    n : int
        The desired random seed. Default: 0.
    """
    random.seed(n)
    np.random.seed(n)


def indent(s, n_spaces=4):
    """Indent a string.

    Parameters
    ----------
    s : str
        The string
    n_spaces : int
        Number of spaces to put at the beginning of each line.

    Returns
    -------
    str

    Examples
    --------
        >>> s = 'First line\\nSecond line'
        >>> print(s)
        First line
        Second line
        >>> print(indent(s, 4))
            First line
            Second line
    """
    return '    ' + s.replace('\n', '\n    ')



from svvamp import RuleApproval, Profile


def test_approval_threshold_setter():
    """
        >>> rule = RuleApproval()
        >>> rule.approval_threshold = 0
        >>> rule.approval_threshold
        0.0
        >>> rule.approval_threshold = 0.1
        >>> rule.approval_threshold
        0.1
        >>> rule.approval_threshold = 'rosebud'
        Traceback (most recent call last):
        ValueError: Unknown value for approval_threshold: rosebud
    """
    pass


def test_approval_comparator_setter():
    """
        >>> rule = RuleApproval()
        >>> rule.approval_comparator = '>'
        >>> rule.approval_comparator
        '>'
        >>> rule.approval_comparator = '>='
        >>> rule.approval_comparator
        '>='
        >>> rule.approval_comparator = '<'
        Traceback (most recent call last):
        ValueError: Unknown option for approval_comparator: <
    """
    pass


def test_comparator_greater_or_equal():
    """
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleApproval(approval_comparator='>=')(profile)
        >>> rule.w_
        0
    """


def test_tm_um_c():
    """
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleApproval()(profile)
        >>> rule.is_tm_c_(0)
        False
        >>> rule.is_um_c_(0)
        False
    """
    pass



# -*- coding: utf-8 -*-
"""
Created on 13 dec. 2018, 09:04
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.rules.rule import Rule
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile
from svvamp.utils.misc import preferences_ut_to_matrix_duels_ut
from svvamp.utils.pseudo_bool import equal_false
from svvamp.rules.rule_irv import RuleIRV


class RuleCondorcetAbsIRV(Rule):
    """Absolute-Condorcet Instant Runoff Voting.

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleCondorcetAbsIRV()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        scores =
        [2. 0. 1.]
        candidates_by_scores_best_to_worst
        [0 2 1]
        scores_best_to_worst
        [2. 1. 0.]
        w = 0
        score_w = 2.0
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  True
        is_based_on_rk =  True
        is_based_on_ut_minus1_1 =  False
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (False) ==> Condorcet_c_rk (False)     ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (False)     ==>     Condorcet_ut_abs_c (True)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (True)  ==> maj_fav_c_rk (True)        ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (True)  ==> majority_favorite_c_ut (True)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (True)                 ==>                IgnMC_c (True)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (True)                 ==>                InfMC_c (True)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = nan
        log_im: im_option = lazy
        candidates_im =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = nan
        log_um: um_option = lazy
        candidates_um =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = False
        log_icm: icm_option = exact
        candidates_icm =
        [0. 0. 0.]
        necessary_coalition_size_icm =
        [0. 6. 4.]
        sufficient_coalition_size_icm =
        [0. 6. 4.]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = nan
        log_cm: cm_option = fast, tm_option = exact
        candidates_cm =
        [ 0.  0. nan]
        necessary_coalition_size_cm =
        [0. 1. 2.]
        sufficient_coalition_size_cm =
        [0. 2. 4.]

    Notes
    -----
    .. note::

        When in doubt between ``CondorcetAbsIRV`` and :class:`CondorcetVtbIRV`, we suggest to use
        :class:`CondorcetVtbIRV`.

    Each voter must provide a weak order, and a strict total order that is coherent with this weak order (i.e.,
    is a tie-breaking of this weak order).

    If there is a Condorcet winner (computed with the weak orders, i.e. in the sense of
    :attr:`matrix_victories_ut_abs`), then she is elected. Otherwise, :class:`RuleIRV` is used (with the strict total
    orders).

    If sincere preferences are strict total orders, then this voting system is equivalent to :class:`CondorcetVtbIRV`
    for sincere voting, but manipulators have more possibilities (they can pretend to have ties in their
    preferences). In that case especially, it is a more 'natural' framework to use :class:`CondorcetVtbIRV`.

    * :meth:`is_cm_`:

        * :attr:`cm_option` = ``'fast'``: Rely on :class:`RuleIRV`'s fast algorithm. Polynomial heuristic. Can prove CM
          but unable to decide non-CM (except in rare obvious cases).
        * :attr:`cm_option` = ``'slow'``: Rely on :class:`RuleExhaustiveBallot`'s exact algorithm. Non-polynomial
          heuristic (:math:`2^{n_c}`). Quite efficient to prove CM or non-CM.
        * :attr:`cm_option` = ``'very_slow'``: Rely on :class:`RuleIRV`'s exact algorithm. Non-polynomial
          heuristic (:math:`n_c!`). Very efficient to prove CM or non-CM.
        * :attr:`cm_option` = ``'exact'``: Non-polynomial algorithm from superclass :class:`Rule`.

        Each algorithm above exploits the faster ones. For example, if :attr:`cm_option` = ``'very_slow'``, SVVAMP
        tries the fast algorithm first, then the slow one, then the 'very slow' one. As soon as it reaches a
        decision, computation stops.

    * :meth:`is_icm_`: Exact in polynomial time.
    * :meth:`is_im_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_iia_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_tm_`: Exact in polynomial time.
    * :meth:`is_um_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.

    References
    ----------
    'Condorcet criterion, ordinality and reduction of coalitional manipulability', François Durand,
    Fabien Mathieu and Ludovic Noirie, 2014.
    """

    full_name = 'Absolute-Condorcet IRV'
    abbreviation = 'ACIRV'

    options_parameters = Rule.options_parameters.copy()
    options_parameters.update({
        'cm_option': {'allowed': {'fast', 'slow', 'very_slow', 'exact'}, 'default': 'fast'},
        'tm_option': {'allowed': ['exact'], 'default': 'exact'},
        'icm_option': {'allowed': {'exact'}, 'default': 'exact'}
    })

    def __init__(self, **kwargs):
        self.irv_ = None
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_um=False, precheck_icm=False,
            log_identity="CONDORCET_ABS_IRV", **kwargs
        )

    def __call__(self, profile):
        self.delete_cache(suffix='_')
        self.profile_ = profile
        # Grab the IRV ballot of the profile (or create it)
        irv_options = {}
        if self.cm_option == 'fast':
            irv_options['cm_option'] = 'fast'
        elif self.cm_option == 'slow':
            irv_options['cm_option'] = 'slow'
        else:  # self.cm_option in {'very_slow', 'exact'}
            irv_options['cm_option'] = 'exact'
        self.irv_ = RuleIRV(**irv_options)(self.profile_)
        return self

    # %% Counting the ballots

    @cached_property
    def _counts_ballots_(self):
        self.mylog("Count ballots", 1)
        if not np.isnan(self.profile_.condorcet_winner_ut_abs):
            w = self.profile_.condorcet_winner_ut_abs
            scores = np.sum(self.profile_.matrix_victories_ut_abs, 1)
            candidates_by_scores_best_to_worst = np.argsort(- scores, kind='mergesort')
            v_might_be_pivotal = np.zeros(self.profile_.n_v)
            for c in np.where(self.profile_.matrix_duels_ut[w, :] <= self.profile_.n_v / 2 + 1)[0]:
                if c == w:
                    continue
                # Search voters who can prevent the victory for ``w`` against ``c``.
                v_might_be_pivotal[self.profile_.preferences_ut[:, w] > self.profile_.preferences_ut[:, c]] = True
        else:
            eb = self.irv_.eb_
            w = eb.w_
            scores = eb.scores_
            candidates_by_scores_best_to_worst = eb.candidates_by_scores_best_to_worst_
            # First way of being (maybe) pivotal: change the result of IRV.
            v_might_be_pivotal = np.copy(eb.v_might_be_pivotal_)
            # Another way of being (maybe) pivotal: create a Condorcet winner.
            for c in range(self.profile_.n_c):
                if np.any(self.profile_.matrix_duels_ut[c, np.not_equal(np.array(range(self.profile_.n_c)), c)]
                          <= self.profile_.n_v / 2 - 1):
                    # ``c`` cannot become a Condorcet winner.
                    continue
                # ``close_candidates`` are the candidates against which ``c`` does not have a victory.
                close_candidates = np.less_equal(self.profile_.matrix_duels_ut[c, :], self.profile_.n_v / 2)
                close_candidates[c] = False
                # Voter ``v`` can make ``c`` become a Condorcet winner iff, among ``close_candidates``, she likes ``c``
                # the least (this way, she can improve ``c``'s scores in all these duels, compared to her sincere
                # voting).
                v_might_be_pivotal[np.all(np.less(
                    self.profile_.preferences_ut[:, c][:, np.newaxis], self.profile_.preferences_ut[:, close_candidates]
                ), 1)] = True
        return {'w': w, 'scores': scores, 'candidates_by_scores_best_to_worst': candidates_by_scores_best_to_worst,
                'v_might_be_pivotal': v_might_be_pivotal}

    @cached_property
    def w_(self):
        return self._counts_ballots_['w']

    @cached_property
    def scores_(self):
        """1d or 2d array.

            * If there is a Condorcet winner, then ``scores[c]`` is the number of victories for ``c`` in
              ``matrix_victories_ut_abs``.
            * Otherwise, ``scores[r, c]`` is defined like in :class:`RuleIRV`: it is the number of voters who vote for
              candidate ``c`` at round ``r``. For eliminated candidates, ``scores[r, c] = numpy.nan``. In contrast,
              ``scores[r, c] = 0`` means that ``c`` is present at round ``r`` but no voter votes for ``c``.
        """
        return self._counts_ballots_['scores']

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        """1d array of integers.

            * If there is a Condorcet winner, candidates are sorted according to their (scalar) score.
            * Otherwise, ``candidates_by_scores_best_to_worst`` is the list of all candidates in the reverse order of
              their IRV elimination.
        """
        return self._counts_ballots_['candidates_by_scores_best_to_worst']

    @cached_property
    def v_might_im_for_c_(self):
        return np.tile(self._counts_ballots_['v_might_be_pivotal'][:, np.newaxis], (1, self.profile_.n_c))

    # %% Manipulation criteria of the voting system

    @cached_property
    def meets_majority_favorite_c_rk_ctb(self):
        return True

    @cached_property
    def meets_condorcet_c_ut_abs(self):
        return True

    # %% Individual manipulation (IM)

    # TODO: should be implemented.

    # %% Trivial Manipulation (TM)

    # Use the general methods from class Rule.

    # %% Unison manipulation (UM)

    # TODO: should be implemented .

    # %% Ignorant-Coalition Manipulation (ICM)

    # Use the general methods from class Rule.

    # %% Coalition Manipulation (CM)

    @cached_property
    def losing_candidates_(self):
        """If ``irv_.w_`` does not win, then we put her first. Other losers are sorted as usual. (scores in
        ``matrix_duels_ut``).

        Examples
        --------
            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 1, 2],
            ...     [1, 2, 0],
            ...     [2, 1, 0],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleCondorcetAbsIRV()(profile)
            >>> rule.candidates_cm_
            array([nan,  0.,  1.])
        """
        self.mylog("Compute ordered list of losing candidates", 1)
        if self.w_ == self.irv_.w_:
            # As usual
            losing_candidates = np.concatenate((
                np.array(range(0, self.w_), dtype=int), np.array(range(self.w_ + 1, self.profile_.n_c), dtype=int)))
            losing_candidates = losing_candidates[np.argsort(
                - self.profile_.matrix_duels_ut[losing_candidates, self.w_], kind='mergesort')]
        else:
            # Put irv_.w_ first.
            losing_candidates = np.array([
                c for c in range(self.profile_.n_c) if c != self.w_ and c != self.irv_.w_]).astype(np.int)
            losing_candidates = losing_candidates[np.argsort(
                - self.profile_.matrix_duels_ut[losing_candidates, self.w_], kind='mergesort')]
            losing_candidates = np.concatenate(([self.irv_.w_], losing_candidates))
        return losing_candidates

    def _cm_preliminary_checks_general_subclass_(self):
        # We check IRV first.
        if self.w_ != self.irv_.w_:
            # Then IRV is manipulable anyway (for ``w``, so a precheck on IRV would give us no information).
            return
        if self.cm_option != "fast":
            if self.cm_option == "slow":
                self.irv_.cm_option = "slow"
            else:
                self.irv_.cm_option = "exact"
            if equal_false(self.irv_.is_cm_):
                # Condorcification theorem apply.
                self.mylog("CM impossible (since it is impossible for IRV)", 2)
                self._is_cm = False
                self._candidates_cm[:] = False
                self._cm_was_computed_with_candidates = True

    def _cm_main_work_c_(self, c, optimize_bounds):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleCondorcetAbsIRV(cm_option='slow')(profile)
            >>> rule.candidates_cm_
            array([0., 0., 0.])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleCondorcetAbsIRV(cm_option='very_slow')(profile)
            >>> rule.candidates_cm_
            array([0., 0., 0.])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleCondorcetAbsIRV(cm_option='exact')(profile)
            >>> rule.candidates_cm_
            array([0., 0., 0.])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1, 3],
            ...     [1, 0, 2, 3],
            ...     [1, 0, 3, 2],
            ...     [2, 3, 0, 1],
            ...     [3, 0, 1, 2],
            ... ])
            >>> rule = RuleCondorcetAbsIRV(cm_option='slow')(profile)
            >>> rule.is_cm_
            nan

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1, 3],
            ...     [1, 0, 2, 3],
            ...     [1, 0, 3, 2],
            ...     [2, 3, 0, 1],
            ...     [3, 0, 1, 2],
            ... ])
            >>> rule = RuleCondorcetAbsIRV(cm_option='very_slow')(profile)
            >>> rule.is_cm_
            False

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1, 3],
            ...     [1, 2, 0, 3],
            ...     [2, 1, 3, 0],
            ...     [3, 1, 0, 2],
            ...     [3, 2, 1, 0],
            ... ])
            >>> rule = RuleCondorcetAbsIRV(cm_option='exact')(profile)
            >>> rule.is_cm_
            True

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 1, 0],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleCondorcetAbsIRV(cm_option='very_slow')(profile)
            >>> rule.is_cm_c_(1)
            False
        """
        # Decondorcification is independent from what we will do for IRV. It just gives a necessary coalition size.
        #
        # 1) If ``irv_.w_`` is Condorcet:
        #
        #    * Manipulate IRV --> use IRV, we know.
        #    * Prevent ``w`` from being a Condorcet winner --> we know.
        #    * If incomplete: we need ``irv_.cm_`` (incomplete). But if IRV works for ``c`` and Cond-IRV does not,
        #      we might need IRV.CM.complete (we will know after).
        #
        # 2) If ``irv_.w_ != CondIRV.w_`` (which is Condorcet):
        #
        #    a) If ``c = irv_.w_``:
        #
        #       * Make ``w`` win: easy, and we have ``suff = n_m`` (``nec`` can be less)
        #       * Prevent ``CondIRV.w`` from being Condorcet --> we know.
        #
        #   b) If ``c != irv_.w``:
        #
        #      * Make ``c`` win: complicated. We can use the elimination path we know, but it might not succeed.
        #      * Prevent ``CondIRV.w`` from being Condorcet --> we know.
        #      * If incomplete: we do not even need ``irv_.CM``. But we need to make sure that we try ``w`` first in
        #        ``losing_candidates_``. However, if it fails, we will need to try other ``c``'s. What do we do then?
        #        Use a fast IRV on the fly? Or simply try the elimination path suggested by IRV? Etc.  Maybe the
        #        first is better (cheap and likely to be more efficient).
        #
        # 3) If there is no Condorcet winner:
        #
        #    * Manipulate IRV --> use IRV, we know.
        #    * Prevent a Condorcet winner --> trivial.
        #    * If incomplete: cf. 1), in fact it is the same. When do we calculate IRV? It can be after we examine
        #      ``irv_.w_``; so, not at the very beginning
        n_m = self.profile_.matrix_duels_ut[c, self.w_]
        n_s = self.profile_.n_v - n_m
        candidates = np.array(range(self.profile_.n_c))
        preferences_utilities_s = self.profile_.preferences_ut[np.logical_not(self.v_wants_to_help_c_[:, c]), :]
        matrix_duels_temp = (preferences_ut_to_matrix_duels_ut(preferences_utilities_s))
        self.mylogm("CM: matrix_duels_temp =", matrix_duels_temp, 3)
        # More preliminary checks. It's more convenient to put them in that method, because we need
        # ``preferences_utilities_s`` and ``matrix_duels_temp``.
        # ``min_d(matrix_duels_temp[c, d]) + n_m > (n_s + n_m) / 2``, i.e.:
        # ``n_m >= n_s + 1 - 2 * min_d(matrix_duels_temp[c, d])``
        d_neq_c = (np.array(range(self.profile_.n_c)) != c)
        n_manip_becomes_cond = np.maximum(n_s + 1 - 2 * np.min(matrix_duels_temp[c, d_neq_c]), 0)
        self.mylogv("CM: n_manip_becomes_cond =", n_manip_becomes_cond, 3)
        self._update_sufficient(self._sufficient_coalition_size_cm, c, n_manip_becomes_cond,
                                'CM: Update sufficient_coalition_size_cm[c] = n_manip_becomes_cond =')
        if not optimize_bounds and n_m >= self._sufficient_coalition_size_cm[c]:  # pragma: no cover
            # TO DO: Investigate whether this case can actually happen.
            self._reached_uncovered_code()
            return True
        # Prevent another cond. Look at the weakest duel for ``d``, she has ``matrix_duels_temp[d, e]``. We simply need
        # that:
        # ``matrix_duels_temp[d, e] <= (n_s + n_m) / 2``
        # ``2 * max_d(min_e(matrix_duels_temp[d, e])) - n_s <= n_m``
        n_manip_prevent_cond = 0
        for d in candidates[d_neq_c]:
            e_neq_d = (np.array(range(self.profile_.n_c)) != d)
            n_prevent_d = np.maximum(2 * np.min(matrix_duels_temp[d, e_neq_d]) - n_s, 0)
            n_manip_prevent_cond = max(n_manip_prevent_cond, n_prevent_d)
        self.mylogv("CM: n_manip_prevent_cond =", n_manip_prevent_cond, 3)
        self._update_necessary(self._necessary_coalition_size_cm, c, n_manip_prevent_cond,
                               'CM: Update necessary_coalition_size_cm[c] = n_manip_prevent_cond =')
        if not optimize_bounds and self._necessary_coalition_size_cm[c] > n_m:  # pragma: no cover
            # TO DO: Investigate whether this case can actually happen.
            self._reached_uncovered_code()
            return True

        is_quick_escape_one = False
        if self.w_ == self.irv_.w_:
            if self.cm_option == "fast":
                self.irv_.cm_option = "fast"
            elif self.cm_option == "slow":
                self.irv_.cm_option = "slow"
            else:
                self.irv_.cm_option = "exact"
            if optimize_bounds:
                self.irv_.is_cm_c_with_bounds_(c)
            else:
                self.irv_.is_cm_c_(c)
                is_quick_escape_one = True
            self.mylog('CM: c != self.irv_.w_ == self.w_', 3)
            self.mylogv("CM: self.irv_.sufficient_coalition_size_cm_[c] =",
                        self.irv_.sufficient_coalition_size_cm_[c], 3)
            self._update_sufficient(self._sufficient_coalition_size_cm, c,
                                    max(self.irv_.sufficient_coalition_size_cm_[c], n_manip_prevent_cond),
                                    'CM: Update sufficient_coalition_size[c] =')
            self.mylogv("CM: self.irv_.necessary_coalition_size_cm_[c] =", self.irv_.necessary_coalition_size_cm_[c], 3)
            self._update_necessary(
                self._necessary_coalition_size_cm, c,
                min(n_manip_becomes_cond, max(self.irv_.necessary_coalition_size_cm_[c], n_manip_prevent_cond)),
                'CM: Update necessary_coalition_size[c] =')
        else:
            if c == self.irv_.w_:  # pragma: no cover
                # TO DO: Investigate whether this case can actually happen.
                # self._reached_uncovered_code()
                self.mylog('CM: c == self.irv_.w_ != self.w_', 3)
                self.mylogv("CM: sufficient size for IRV (sincere IRV) =", n_m, 3)
                self._update_sufficient(self._sufficient_coalition_size_cm, c, max(n_m, n_manip_prevent_cond),
                                        'CM: Update sufficient_coalition_size[c] =')
            else:
                self.mylog('CM: c, self.w_, self.irv_.w_ all distinct', 3)
                # We do not know how many manipulators can make ``c`` win in IRV (note that it would not be the same
                # set of manipulators in IRV and here)
        # self.mylogv("CM: Preliminary checks.2: necessary_coalition_size_cm[c] =",
        #             self._necessary_coalition_size_cm[c], 3)
        # self.mylogv("CM: Preliminary checks.2: sufficient_coalition_size_cm[c] =",
        #             self._sufficient_coalition_size_cm[c], 3)
        # self.mylogv("CM: Preliminary checks.2: n_m =", self.profile_.matrix_duels_ut[c, self.w], 3)
        # if self._necessary_coalition_size_cm[c] > n_m:
        #     self.mylogv("CM: Preliminary checks.2: CM is False for c =", c, 2)
        # elif self._sufficient_coalition_size_cm[c] <= n_m:
        #     self.mylogv("CM: Preliminary checks.2: CM is True for c =", c, 2)
        # else:
        #     self.mylogv("CM: Preliminary checks.2: CM is unknown for c =", c, 2)
        # Real work
        is_quick_escape_two = False
        if self.cm_option == 'exact':
            is_quick_escape_two = self._cm_main_work_c_exact_(c, optimize_bounds)
        return is_quick_escape_one or is_quick_escape_two



# -*- coding: utf-8 -*-
"""
Created on 18 jul. 2021, 15:41
Copyright François Durand 2014-2021
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from math import floor, ceil
from svvamp.rules.rule import Rule
from svvamp.utils import type_checker
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile
from svvamp.utils.misc import preferences_ut_to_matrix_duels_ut


class RuleSTAR(Rule):
    """STAR (Scoring Then Automatic Runoff).

    Parameters
    ----------
    min_grade : number
        Minimal grade allowed.
    max_grade : number
        Maximal grade allowed.
    step_grade : number
        Interval between two consecutive allowed grades.

        * If ``step_grade = 0``, all grades in the interval [:attr:`min_grade`, :attr:`max_grade`] are allowed
          ('continuous' set of grades).
        * If ``step_grade > 0``, authorized grades are the multiples of :attr:`step_grade` lying in the interval
          [:attr:`min_grade`, :attr:`max_grade`]. In addition, the grades :attr:`min_grade` and :attr:`max_grade` are
          always authorized, even if they are not multiples of :attr:`step_grade`.

    rescale_grades : bool
        Whether sincere voters rescale their utilities to produce grades.

        * If ``rescale_grades`` = ``True``, then each sincere voter ``v`` applies an affine transformation to send
          her utilities into the interval [:attr:`min_grade`, :attr:`max_grade`].
        * If ``rescale_grades`` = ``False``, then each sincere voter ``v`` clips her utilities into the interval
          [:attr:`min_grade`, :attr:`max_grade`].

        See :attr:`ballots` for more details.

    Examples
    --------
    STAR does not :attr:`meets_majority_favorite_c_ut`:

        >>> profile = Profile(preferences_ut=[
        ...     [ 0. ,  0. ,  0. , -1. , -0.5,  0. ],
        ...     [ 0.5,  0.5,  0.5,  1. ,  0. , -1. ],
        ...     [ 0. , -0.5, -1. ,  1. , -1. , -1. ],
        ... ], preferences_rk=[
        ...     [2, 0, 5, 1, 4, 3],
        ...     [3, 0, 1, 2, 4, 5],
        ...     [3, 0, 1, 4, 5, 2],
        ... ])
        >>> RuleSTAR()(profile).w_
        0
        >>> profile.majority_favorite_ut
        3

    STAR does not ::

    Notes
    -----
    Each voter attributes a grade to each candidate. By default, authorized grades are all numbers in the interval
    [:attr:`min_grade`, :attr:`max_grade`]. To use a discrete set of notes, modify attribute :attr:`step_grade`.

    The two candidates with highest average grades are selected for a virtual second round. During this second round,
    the two candidates are compared by majority voting: the score of one candidate is the number of voters who
    gave her a strictly larger grade than to the other candidate.

    Default behavior of sincere voters: voter ``v`` applies an affine transformation to her utilities
    :attr:`preferences_ut`\ ``[v, :]`` to get her grades, such that her least-liked candidate receives
    :attr:`min_grade` and her most-liked candidate receives :attr:`max_grade`. To modify this behavior, use attribute
    :attr:`rescale_grades`. For more details about the behavior of sincere voters, see :attr:`ballots`.

    * :meth:`is_cm_`, :meth:`is_icm_`, :meth:`is_im_`, :meth:`is_tm_`, :meth:`is_um_`: Exact in polynomial time.
    """

    full_name = 'Scoring Then Automatic Runoff'
    abbreviation = 'STAR'

    options_parameters = Rule.options_parameters.copy()
    options_parameters.update({
        'max_grade': {'allowed': np.isfinite, 'default': 1},
        'min_grade': {'allowed': np.isfinite, 'default': 0},
        'step_grade': {'allowed': np.isfinite, 'default': 0},
        'rescale_grades': {'allowed': type_checker.is_bool, 'default': True},
        'im_option': {'allowed': ['exact'], 'default': 'exact'},
        'tm_option': {'allowed': ['exact'], 'default': 'exact'},
        'um_option': {'allowed': ['exact'], 'default': 'exact'},
        'icm_option': {'allowed': ['exact'], 'default': 'exact'},
        'cm_option': {'allowed': ['exact'], 'default': 'exact'}
    })

    def __init__(self, **kwargs):
        self._min_grade = None
        self._max_grade = None
        self._step_grade = None
        self._rescale_grades = None
        super().__init__(
            with_two_candidates_reduces_to_plurality=False,
            # # Even if ``rescale_grades = True``, a voter who has the same utility for ``c`` and ``d`` will not vote
            # # the same in STAR and in Plurality.
            is_based_on_rk=False,
            precheck_um=False, precheck_tm=False, precheck_icm=False,
            log_identity="STAR", **kwargs
        )

    # %% Setting the parameters

    @property
    def min_grade(self):
        return self._min_grade

    @min_grade.setter
    def min_grade(self, value):
        if self._min_grade == value:
            return
        if self.options_parameters['min_grade']['allowed'](value):
            self.mylogv("Setting min_grade =", value, 1)
            self._min_grade = value
            self._result_options['min_grade'] = value
            self.delete_cache()
        else:
            raise ValueError("Unknown value for min_grade: " + format(value))

    @property
    def max_grade(self):
        return self._max_grade

    @max_grade.setter
    def max_grade(self, value):
        if self._max_grade == value:
            return
        if self.options_parameters['max_grade']['allowed'](value):
            self.mylogv("Setting max_grade =", value, 1)
            self._max_grade = value
            self._result_options['max_grade'] = value
            self.delete_cache()
        else:
            raise ValueError("Unknown value for max_grade: " + format(value))

    @property
    def step_grade(self):
        return self._step_grade

    @step_grade.setter
    def step_grade(self, value):
        if self._step_grade == value:
            return
        if self.options_parameters['step_grade']['allowed'](value):
            self.mylogv("Setting step_grade =", value, 1)
            self._step_grade = value
            self._result_options['step_grade'] = value
            self.delete_cache()
        else:
            raise ValueError("Unknown value for step_grade: " + format(value))

    @property
    def rescale_grades(self):
        return self._rescale_grades

    @rescale_grades.setter
    def rescale_grades(self, value):
        if self._rescale_grades == value:
            return
        if self.options_parameters['rescale_grades']['allowed'](value):
            self.mylogv("Setting rescale_grades =", value, 1)
            self._rescale_grades = value
            self._result_options['rescale_grades'] = value
            self.delete_cache()
        else:
            raise ValueError("Unknown value for rescale_grades: " + format(value))

    @cached_property
    def allowed_grades(self):
        """List or None. If ``step_grade`` is positive, the list of authorized grades. If ``step_grade`` is zero
        (continuous set of grades), then None.
        """
        if self.step_grade == 0:
            return None
        else:
            i_lowest_rung = floor(self.min_grade / self.step_grade) + 1
            i_highest_rung = ceil(self.max_grade / self.step_grade) - 1
            return np.concatenate((
                [self.min_grade],
                np.array(range(i_lowest_rung, i_highest_rung + 1)) * self.step_grade,
                [self.max_grade]
            ))

    @cached_property
    def second_best_grade(self):
        """Number : second highest possible grade. Useful for manipulation algorithms.

        If step_grade = 0, then conventionally, this returns `max_grade` (whereas the actual second best possible
        grade is `max_grade` - epsilon, for any positive epsilon).
        """
        if self.step_grade == 0:
            return self.max_grade
        else:
            return self.allowed_grades[-2]

    # %% Counting the ballots

    @cached_property
    def ballots_(self):
        """2d array of integers. ``ballots[v, c]`` is the grade attributed by voter ``v`` to candidate ``c`` (when
        voting sincerely). The following process is used.

        1. Convert utilities into grades in the interval [:attr:`min_grade`, :attr:`max_grade`].

            * If :attr:`rescale_grades` = ``True``, then each voter ``v`` applies an affine transformation to
              :attr:`preferences_ut`\ ``[v, :]`` such that her least-liked candidate receives :attr:`min_grade` and
              her most-liked candidate receives :attr:`max_grade`. Exception: if she is indifferent between all
              candidates, then she attributes (:attr:`min_grade` + :attr:`max_grade`) / 2 to all of them.

            * If :attr:`rescale_grades` = ``False``, then each voter ``v`` clips her utilities into the interval
              [:attr:`min_grade`, :attr:`max_grade`]: each utility greater than :attr:`max_grade` (resp. lower than
              :attr:`min_grade`) becomes :attr:`max_grade` (resp. :attr:`min_grade`).

        2. If :attr:`step_grades` > 0 (discrete set of grades), round each grade to the closest authorized grade.
        """
        self.mylog("Compute ballots", 1)
        # Rescale (or not)
        if self.rescale_grades:
            # Multiplicative renormalization
            max_util = np.max(self.profile_.preferences_ut, axis=1)
            min_util = np.min(self.profile_.preferences_ut, axis=1)
            delta_util = max_util - min_util
            ballots = np.divide(
                self.profile_.preferences_ut * (self.max_grade - self.min_grade),
                delta_util[:, np.newaxis],
                out=np.zeros(self.profile_.preferences_ut.shape),
                where=delta_util[:, np.newaxis] != 0
            )  # `out` and `where` ensure that when dividing by zero, the result will be zero.
            # Additive renormalization
            middle_grade = (np.max(ballots, axis=1) + np.min(ballots, axis=1)) / 2
            ballots += (self.max_grade - self.min_grade) / 2 - middle_grade[:, np.newaxis]
        else:
            ballots = np.clip(self.profile_.preferences_ut, self.min_grade, self.max_grade)
        # Round (or not)
        if self.step_grade != 0:
            if self.min_grade % 1 == 0 and self.max_grade % 1 == 0 and self.step_grade % 1 == 0:
                # There is a faster version for this (common) use case.
                ballots = np.array(np.rint(ballots), dtype=int)
            else:
                frontiers = (self.allowed_grades[0:-1] + self.allowed_grades[1:]) / 2
                for v in range(self.profile_.n_v):
                    ballots[v, :] = self.allowed_grades[np.digitize(ballots[v, :], frontiers)]
        return ballots

    @cached_property
    def _count_ballots_(self):
        self.mylog("Count ballots", 1)
        # First round
        scores_first_round = np.sum(self.ballots_, 0)
        c = np.argmax(scores_first_round)
        d = np.argmax([-np.inf if i == c else s for i, s in enumerate(scores_first_round)])
        # Second round
        scores_second_round = np.zeros(self.profile_.n_c)
        scores_second_round[c] = np.sum(self.ballots_[:, c] > self.ballots_[:, d])
        scores_second_round[d] = np.sum(self.ballots_[:, d] > self.ballots_[:, c])
        w = c if scores_second_round[c] > scores_second_round[d] else d
        return {'w': w, 'selected_one': c, 'selected_two': d,
                'scores': np.array([scores_first_round, scores_second_round])}

    @cached_property
    def w_(self):
        return self._count_ballots_['w']

    @cached_property
    def scores_(self):
        """2d array of numbers. ``scores_[0, c]`` is the total grade of candidate ``c``. ``scores_[1, c]`` the number
        of voters who vote for ``c`` during the second round (and 0 if ``c`` is not selected for the second round).
        """
        return self._count_ballots_['scores']

    @cached_property
    def scores_average_(self):
        """2d array of floats. ``scores_[0, c]`` is the average grade of candidate ``c``. ``scores_[1, c]`` the number
        of voters who vote for ``c`` during the second round (and 0 if ``c`` is not selected for the second round).
        """
        return np.array([
            self.scores_[0, :] / self.profile_.n_v,
            self.scores_[1, :]
        ])

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        """1d array of integers. ``candidates_by_scores_best_to_worst[k]`` is the candidate with ``k``\ :sup:`th`
        best score. Finalists are sorted by their score at second round. Other candidates are sorted by their score
        at first round.
        """
        self.mylog("Count candidates_by_scores_best_to_worst", 1)
        return np.argsort(- np.max(self.scores_, 0))

    # %% Manipulation criteria of the voting system

    @cached_property
    def meets_ignmc_c_ctb(self):
        return True

    # %% Individual manipulation (IM)

    def _im_main_work_v_(self, v, c_is_wanted, nb_wanted_undecided, stop_if_true):
        other_voters = np.ones(self.profile_.n_v, dtype=bool)
        other_voters[v] = False
        scores_first_round_without_v = self.scores_[0, :] - self.ballots_[v, :]
        matrix_duels_without_v = preferences_ut_to_matrix_duels_ut(self.ballots_[other_voters, :])
        for c in range(self.profile_.n_c):
            if not c_is_wanted[c]:
                continue
            if not np.isneginf(self._v_im_for_c[v, c]):
                continue
            scores_first_round_c = np.around(scores_first_round_without_v[c] + self.max_grade, 12)
            scores_first_round_s2 = np.around(scores_first_round_without_v[c] + self.min_grade, 12)
            if scores_first_round_c < scores_first_round_s2:
                # Then c cannot go to second round
                self._v_im_for_c[v, c] = False
            else:
                # Case 1: c in {selected_one_without_v, selected_two_without_v} ==> c can go to second round.
                # Case 2: c not in {s1, s2} ==> since c can do better than s2, c can go to second round.
                # => Anyway, c might go to second round.
                n_m_needed_second_round_against_d = np.maximum(
                    matrix_duels_without_v[:, c] - matrix_duels_without_v[c, :]
                    + (np.array(range(self.profile_.n_c)) < c),
                    0
                )
                n_m_needed_second_round_against_d[c] = np.nan
                manipulation_found = False
                # First try
                d_loses_already_against_c = np.where(n_m_needed_second_round_against_d == 0)[0]
                d = max(d_loses_already_against_c, key=scores_first_round_without_v.__getitem__)
                scores_first_round = scores_first_round_without_v.copy()
                scores_first_round[c] += self.max_grade - self.min_grade
                scores_first_round[d] += self.max_grade - self.min_grade
                selected_one = np.argmax(scores_first_round)
                selected_two = np.argmax([-np.inf if i == selected_one else s
                                          for i, s in enumerate(scores_first_round)])
                if {selected_one, selected_two} == {c, d}:
                    manipulation_found = True
                else:
                    # Second try
                    d_can_lose_against_c_by_manipulation = np.where(n_m_needed_second_round_against_d == 1)[0]
                    d = max(d_can_lose_against_c_by_manipulation, key=scores_first_round_without_v.__getitem__)
                    scores_first_round = scores_first_round_without_v.copy()
                    scores_first_round[c] += self.max_grade - self.min_grade
                    scores_first_round[d] += self.second_best_grade - self.min_grade  # Minus epsilon if step_grade == 0
                    selected_one = np.argmax(scores_first_round)
                    selected_two = np.argmax([-np.inf if i == selected_one else s
                                              for i, s in enumerate(scores_first_round)])
                    if {selected_one, selected_two} == {c, d}:
                        manipulation_found = True
                    # ... but there is an exception:
                    if self.step_grade == 0 and scores_first_round[d] == scores_first_round[selected_two]:
                        selected_three = np.argmax([-np.inf if i in {selected_one, selected_two} else s
                                                    for i, s in enumerate(scores_first_round)])
                        if scores_first_round[selected_three] == scores_first_round[d]:
                            manipulation_found = False
                if manipulation_found:
                    self._v_im_for_c[v, c] = True
                    self._candidates_im[c] = True
                    self._voters_im[v] = True
                    self._is_im = True
                    self.mylog("IM found", 3)
                    if stop_if_true:
                        return
                else:
                    self._v_im_for_c[v, c] = False
                nb_wanted_undecided -= 1
                if nb_wanted_undecided == 0:
                    return

    # %% Coalition Manipulation (CM)

    def _cm_main_work_c_(self, c, optimize_bounds):
        ballots_s = self.ballots_[np.logical_not(self.v_wants_to_help_c_[:, c]), :]
        scores_first_round_s = np.sum(ballots_s, 0)
        for d in range(self.profile_.n_c):
            # We want to put c and d at second round, and make c win.
            if d == c:
                continue
            # How many manipulator to ensure that c wins the second round?
            score_second_round_c_d = np.sum(ballots_s[:, c] > ballots_s[:, d])
            score_second_round_d_c = np.sum(ballots_s[:, d] > ballots_s[:, c])
            n_2 = max(score_second_round_d_c - score_second_round_c_d + int(d < c), 0)
            # Consequences on first round
            scores_first_round = scores_first_round_s.copy().astype(float)
            scores_first_round[c] += n_2 * self.max_grade
            scores_first_round[d] += n_2 * self.second_best_grade  # Exactly if step_grade > 0, minus epsilon if == 0
            if self.profile_.n_c == 2:
                self._update_sufficient(self._sufficient_coalition_size_cm, c, n_2,
                                        'CM: Update sufficient_coalition_size_cm[c] = n_m =')
                continue
            # How many additional manipulators to win the first round?
            score_c = scores_first_round[c]
            score_d = scores_first_round[d]
            scores_first_round[c] = -np.inf
            scores_first_round[d] = -np.inf
            best_challenger = np.argmax(scores_first_round)
            score_challenger = scores_first_round[best_challenger]
            n_c_not_rounded = np.around((score_challenger - score_c) / (self.max_grade - self.min_grade), 12)
            if n_c_not_rounded % 1 == 0 and c > best_challenger:
                n_c = n_c_not_rounded + 1
            else:
                n_c = ceil(n_c_not_rounded)
            n_d_not_rounded = np.around((score_challenger - score_d) / (self.max_grade - self.min_grade), 12)
            if n_d_not_rounded % 1 == 0 and ((self.step_grade == 0 and n_2 > 0) or d > best_challenger):
                n_d = n_d_not_rounded + 1
            else:
                n_d = ceil(n_d_not_rounded)
            n_1 = max(n_c, n_d, 0)
            # Total number of manipulators
            self._update_sufficient(self._sufficient_coalition_size_cm, c, n_1 + n_2,
                                    'CM: Update sufficient_coalition_size_cm[c] = n_m =')
        # Final conclusion: we know that this algorithm is exact.
        self._necessary_coalition_size_cm[c] = self._sufficient_coalition_size_cm[c]

    # %% Trivial Manipulation (TM)

    def _tm_main_work_c_(self, c):
        # Manipulators give min_grade to all candidates, except max_grade for c.
        preferences_test = np.copy(self.profile_.preferences_ut)
        preferences_test[self.v_wants_to_help_c_[:, c], :] = self.min_grade
        preferences_test[self.v_wants_to_help_c_[:, c], c] = self.max_grade
        w_test = self._copy(profile=Profile(preferences_ut=preferences_test)).w_
        self._candidates_tm[c] = (w_test == c)

    # %% Unison Manipulation (UM)

    def _um_main_work_c_(self, c):
        n_m = self.profile_.matrix_duels_ut[c, self.w_]
        ballots_s = self.ballots_[np.logical_not(self.v_wants_to_help_c_[:, c]), :]
        scores_first_round_s = np.sum(ballots_s, 0)
        self._candidates_um[c] = False  # until otherwise proven
        for d in range(self.profile_.n_c):
            # We want to put c and d at second round, and make c win.
            if d == c:
                continue
            # How many manipulator to ensure that c wins the second round?
            score_second_round_c_d = np.sum(ballots_s[:, c] > ballots_s[:, d])
            score_second_round_d_c = np.sum(ballots_s[:, d] > ballots_s[:, c])
            n_2 = max(score_second_round_d_c - score_second_round_c_d + int(d < c), 0)
            # Consequences on first round
            scores_first_round = scores_first_round_s.copy().astype(float)
            scores_first_round[c] += n_2 * self.max_grade
            scores_first_round[d] += n_2 * self.second_best_grade  # Exactly if step_grade > 0, minus epsilon if == 0
            if self.profile_.n_c == 2:
                if n_2 <= n_m:
                    self._candidates_um[c] = True
                    return
            # How many additional manipulators to win the first round?
            score_c = scores_first_round[c]
            score_d = scores_first_round[d]
            scores_first_round[c] = -np.inf
            scores_first_round[d] = -np.inf
            best_challenger = np.argmax(scores_first_round)
            score_challenger = scores_first_round[best_challenger]
            n_c_not_rounded = np.around((score_challenger - score_c) / (self.max_grade - self.min_grade), 12)
            if n_c_not_rounded % 1 == 0 and c > best_challenger:
                n_c = n_c_not_rounded + 1
            else:
                n_c = ceil(n_c_not_rounded)
            # N.B.: the following "if" clause is the main difference with CM algorithm.
            if n_2 > 0:
                # Due to UM, all manipulators need to put second-best grade to `d`.
                n_d_not_rounded = np.around((score_challenger - score_d) / (self.second_best_grade - self.min_grade),
                                            12)
            else:
                # No constraint on `d`'s grade, so we can use max_grade.
                n_d_not_rounded = np.around((score_challenger - score_d) / (self.max_grade - self.min_grade),
                                            12)
            if n_d_not_rounded % 1 == 0 and ((self.step_grade == 0 and n_2 > 0) or d > best_challenger):
                n_d = n_d_not_rounded + 1
            else:
                n_d = ceil(n_d_not_rounded)
            n_1 = max(n_c, n_d, 0)
            # Total number of manipulators
            if n_1 + n_2 <= n_m:
                self._candidates_um[c] = True
                return

    # %% Ignorant-Coalition Manipulation (ICM)

    # Since STAR meets IgnMC_c_ctb, the general methods are exact.



# -*- coding: utf-8 -*-
"""
Created on 10 dec. 2018, 15:11
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
import networkx as nx
from svvamp.rules.rule import Rule
from svvamp.utils.misc import preferences_ut_to_matrix_duels_ut
from svvamp.utils.util_cache import cached_property
from svvamp.utils.pseudo_bool import equal_true
from svvamp.preferences.profile import Profile


class RuleMaximin(Rule):
    """Maximin method.

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleMaximin()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        scores =
        [3 2 2]
        candidates_by_scores_best_to_worst
        [0 1 2]
        scores_best_to_worst
        [3 2 2]
        w = 0
        score_w = 3
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  True
        is_based_on_rk =  True
        is_based_on_ut_minus1_1 =  False
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (True)  ==> Condorcet_c_rk (True)      ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (True)      ==>     Condorcet_ut_abs_c (True)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (True)  ==> maj_fav_c_rk (True)        ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (True)  ==> majority_favorite_c_ut (True)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (True)                 ==>                IgnMC_c (True)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (True)                 ==>                InfMC_c (True)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = False
        log_im: im_option = exact
        candidates_im =
        [0. 0. 0.]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = False
        log_um: um_option = exact
        candidates_um =
        [0. 0. 0.]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = False
        log_icm: icm_option = exact
        candidates_icm =
        [0. 0. 0.]
        necessary_coalition_size_icm =
        [0. 6. 4.]
        sufficient_coalition_size_icm =
        [0. 6. 4.]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = False
        log_cm: cm_option = fast, um_option = exact, tm_option = exact
        candidates_cm =
        [0. 0. 0.]
        necessary_coalition_size_cm =
        [0. 2. 3.]
        sufficient_coalition_size_cm =
        [0. 2. 3.]

    Notes
    -----
    Candidate ``c``'s score is the minimum of the row :attr:`matrix_duels_rk`\ ``[c, :]`` (except the diagonal term),
    i.e. the result of candidate ``c`` for her worst duel. The candidate with highest score is declared the winner.
    In case of a tie, the candidate with lowest index wins.

    This method meets the Condorcet criterion.

    * :meth:`is_cm_`: Deciding CM is NP-complete, even for 2 manipulators.

        * :attr:`cm_option` = ``'fast'``: Zuckerman et al. (2011). This approximation algorithm is polynomial and has
          a multiplicative factor of error of 5/3 on the number of manipulators needed.
        * :attr:`cm_option` = ``'exact'``: Non-polynomial algorithm from superclass :class:`Rule`.

    * :meth:`is_icm_`: Exact in polynomial time.
    * :meth:`is_im_`: Exact in polynomial time.
    * :meth:`is_iia_`: Exact in polynomial time.
    * :meth:`is_tm_`: Exact in polynomial time.
    * :meth:`is_um_`: Exact in polynomial time.

    References
    ----------
    'Complexity of Unweighted Coalitional Manipulation under Some Common Voting Rules', Lirong Xia et al., 2009.

    'An algorithm for the coalitional manipulation problem under Maximin', Michael Zuckerman, Omer Lev and
    Jeffrey S. Rosenschein, 2011.
    """

    full_name = 'Maximin'
    abbreviation = 'Max'

    options_parameters = Rule.options_parameters.copy()
    options_parameters.update({
        'im_option': {'allowed': ['exact'], 'default': 'exact'},
        'tm_option': {'allowed': ['exact'], 'default': 'exact'},
        'um_option': {'allowed': ['exact'], 'default': 'exact'},
        'icm_option': {'allowed': ['exact'], 'default': 'exact'},
        'cm_option': {'allowed': ['fast', 'exact'], 'default': 'fast'}
    })

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_icm=False,
            log_identity="MAXIMIN", **kwargs
        )

    @cached_property
    def scores_(self):
        """1d array of integers. ``scores[c]`` is the minimum of the row :attr:`matrix_duels_rk`\ ``[c, :]`` (except
        the diagonal term), i.e. the result of candidate ``c`` for her worst duel.
        """
        self.mylog("Compute scores", 1)
        return np.add(self.profile_.n_v, - np.max(self.profile_.matrix_duels_rk, 0))

    # %% Manipulation criteria of the voting system

    @cached_property
    def meets_condorcet_c_rk_ctb(self):
        return True

    # %% Manipulation routine

    def _vote_strategically_(self, matrix_duels_r, scores_r, c, weight=1):
        """Strategic vote by one elector, in favor of candidate ``c``.

        Parameters
        ----------
        matrix_duels_r : list of list (or ndarray)
            2d array of integers. Diagonal coefficients must have an arbitrary value that is greater than all the
            other ones (``+ inf`` is a good idea).
        scores_r : ndarray
            1d array of integers. Scores corresponding to ``matrix_duels_r``.
        c : int
            Candidate.
        weight : int
            Weight of the voter (used for UM, 1 otherwise).

        Returns
        -------
        None
            This method modifies ``matrix_duels_r`` and ``scores_r`` IN PLACE.

        References
        ----------
        Algorithm from Zuckerman et al. : An Algorithm for the Coalitional Manipulation Problem under Maximin. For
        only one elector (or unison manipulation), this algorithm is optimal.
        """
        self.mylogv("AUX: c =", c, 3)
        self.mylogm("AUX: matrix_duels_r =", matrix_duels_r, 3)
        self.mylogv("AUX: scores_r =", scores_r, 3)
        # Manipulator's vote (denoted P_i in the paper) is represented in Borda format.
        strategic_ballot = np.zeros(self.profile_.n_c)
        strategic_ballot[c] = self.profile_.n_c
        next_borda_score_to_add = self.profile_.n_c - 1
        has_been_sent_to_stacks = np.zeros(self.profile_.n_c)
        candidates = np.array(range(self.profile_.n_c))
        stacks_q = StackFamily()

        # Building the directed graph (cf. paper).
        digraph = np.zeros((self.profile_.n_c, self.profile_.n_c))
        for x in range(self.profile_.n_c):
            if x == c:
                continue
            if matrix_duels_r[x, c] == scores_r[x]:
                continue
            for y in range(self.profile_.n_c):
                if y == c or y == x:
                    continue
                if matrix_duels_r[x, y] == scores_r[x]:
                    digraph[x, y] = True

        while next_borda_score_to_add > 0:
            self.mylogm("AUX: digraph =", digraph, 3)
            self.mylogv("AUX: strategic_ballot =", strategic_ballot, 3)
            # candidates_not_dangerous = set A in the paper
            candidates_not_dangerous = np.logical_and(np.logical_not(np.any(digraph, 1)),
                                                      np.logical_not(strategic_ballot))
            self.mylogv("AUX: candidates_not_dangerous =", candidates_not_dangerous, 3)
            if np.any(candidates_not_dangerous):
                # Transfer these candidates into the stacks
                for score, candidate in zip(scores_r[candidates_not_dangerous], candidates[candidates_not_dangerous]):
                    if not has_been_sent_to_stacks[candidate]:
                        stacks_q.push_front(score, candidate)
                        has_been_sent_to_stacks[candidate] = True
                self.mylogv("AUX: stacks_q =", stacks_q.data, 3)
                # Put a non-dangerous candidate in the ballot
                b = stacks_q.pop_front()
                self.mylogv("AUX: Found non-dangerous candidate b =", b, 3)
                self.mylogv("AUX: stacks_q =", stacks_q.data, 3)
                strategic_ballot[b] = next_borda_score_to_add
                next_borda_score_to_add -= 1
            else:
                s = np.min(scores_r[np.logical_not(strategic_ballot)])
                b_have_s = np.logical_and(np.logical_not(strategic_ballot), scores_r == s)
                self.mylogv("AUX: s =", s, 3)
                self.mylogv("AUX: b_have_s =", b_have_s, 3)
                # In the general case, we take any ``b`` with least score. We choose the one with highest index.
                b = np.where(b_have_s)[0][-1]
                self.mylogv("AUX: First guess b =", b, 3)
                # But there might be a particular case (special cycle)...
                if sum(b_have_s) > 1:
                    found_special_b = False
                    dig = nx.DiGraph(digraph)
                    self.mylogv("AUX: G.nodes = ", dig.nodes, 3)
                    self.mylogv("AUX: G.edges = ", dig.edges, 3)
                    # We look for ``b`` in reverse order of index. This way, if several candidates meet the condition,
                    # the one with highest index is chosen.
                    for b_test in candidates[b_have_s][::-1]:
                        possible_cs = np.where(np.logical_and(digraph[:, b_test], b_have_s))
                        for c_test in possible_cs[0]:
                            self.mylogv("AUX: b_test = ", b_test, 3)
                            self.mylogv("AUX: c_test = ", c_test, 3)
                            if nx.has_path(dig, b_test, c_test):
                                self.mylogv("AUX: Found special b =", b_test, 3)
                                b = b_test
                                found_special_b = True
                                break
                        if found_special_b:
                            break
                # We put ``b`` in the ballot
                strategic_ballot[b] = next_borda_score_to_add
                next_borda_score_to_add -= 1
            # Now update the digraph
            self.mylogm("AUX: digraph =", digraph, 3)
            for y in candidates[np.logical_not(strategic_ballot)]:
                if digraph[y, b] and not has_been_sent_to_stacks[y]:
                    self.mylogv("AUX: y =", y, 3)
                    self.mylogv("AUX: b =", b, 3)
                    self.mylogv("AUX: Transfer from digraph to stack: y =", y, 3)
                    digraph[y, :] = False
                    stacks_q.push_front(scores_r[y], y)
                    has_been_sent_to_stacks[y] = True
                    self.mylogv("AUX: stacks_q =", stacks_q.data, 3)
        # Finally, update ``matrix_duels_r`` and ``scores_r``.
        self.mylogm("AUX: matrix_duels_r =", matrix_duels_r, 3)
        self.mylogv("AUX: scores_r =", scores_r, 3)
        self.mylogv("AUX: strategic_ballot =", strategic_ballot, 3)
        for x in range(self.profile_.n_c):
            for y in range(x + 1, self.profile_.n_c):
                if strategic_ballot[x] > strategic_ballot[y]:
                    matrix_duels_r[x, y] += weight
                else:
                    matrix_duels_r[y, x] += weight
        # The ugly thing below (instead of ``scores_r = np.min(matrix_duels_r, 1)``) is made so that ``scores_r`` is
        # modified in place.
        scores_r_temp = np.min(matrix_duels_r, 1)
        for x in range(self.profile_.n_c):
            scores_r[x] = scores_r_temp[x]
        self.mylogm("AUX: matrix_duels_r =", matrix_duels_r, 3)
        self.mylogv("AUX: scores_r =", scores_r, 3)

    # %% Individual manipulation (IM)

    def _im_main_work_v_(self, v, c_is_wanted, nb_wanted_undecided, stop_if_true):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleMaximin()(profile)
            >>> rule.is_im_c_with_voters_(2)
            (False, array([0., 0., 0., 0., 0.]))

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleMaximin()(profile)
            >>> rule.is_im_
            False

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [1, 0, 2],
            ...     [1, 2, 0],
            ...     [2, 0, 1],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleMaximin()(profile)
            >>> rule.is_im_
            True

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 1, 2],
            ...     [1, 2, 0],
            ...     [2, 1, 0],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleMaximin()(profile)
            >>> rule.is_im_
            False

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1, 3],
            ...     [1, 3, 0, 2],
            ...     [2, 1, 0, 3],
            ...     [2, 1, 3, 0],
            ...     [3, 0, 1, 2],
            ... ])
            >>> rule = RuleMaximin()(profile)
            >>> rule.is_im_
            True
        """
        for c in self.losing_candidates_:
            if not c_is_wanted[c]:
                continue
            if not np.isneginf(self._v_im_for_c[v, c]):
                continue
            self.mylogv("IM: c =", c, 3)
            nb_wanted_undecided -= 1

            matrix_duels_r = np.copy(self.profile_.matrix_duels_ut).astype(np.float)
            for x in range(self.profile_.n_c):
                matrix_duels_r[x, x] = np.inf
                for y in range(x + 1, self.profile_.n_c):
                    if self.profile_.preferences_borda_rk[v, x] > self.profile_.preferences_borda_rk[v, y]:
                        matrix_duels_r[x, y] -= 1
                    else:
                        matrix_duels_r[y, x] -= 1
            scores_r = np.min(matrix_duels_r, 1)

            w_temp = np.argmax(scores_r)
            if w_temp == c:
                self.mylogv("IM: scores_r =", scores_r, 3)
                self.mylog("IM: Manipulation easy (c wins without manipulator's vote)", 3)
                self._v_im_for_c[v, c] = True
                self._candidates_im[c] = True
                self._voters_im[v] = True
                self._is_im = True
                self.mylog("IM found", 3)
                if stop_if_true or nb_wanted_undecided == 0:
                    return
                continue  # pragma: no cover - Not really "executed" because of Python optimization
            # Best we can do: improve ``c`` by one, do not move the others
            if scores_r[w_temp] > scores_r[c] + 1 or (scores_r[w_temp] >= scores_r[c] + 1 and w_temp < c):
                self.mylogv("IM: scores_r =", scores_r, 3)
                self.mylog("IM: Manipulation impossible (score difference is too high)", 3)
                self._v_im_for_c[v, c] = False
                if nb_wanted_undecided == 0:
                    return
                continue  # pragma: no cover - Not really "executed" because of Python optimization

            self._vote_strategically_(matrix_duels_r, scores_r, c, 1)
            w_r = np.argmax(scores_r)
            self.mylogv("IM: w_r =", w_r, 3)
            if w_r == c:
                self.mylog("IM: Manipulation worked!", 3)
            else:
                self.mylog("IM: Manipulation failed...", 3)

            # We can conclude.
            self._v_im_for_c[v, c] = (w_r == c)
            if equal_true(self._v_im_for_c[v, c]):
                self._candidates_im[c] = True
                self._voters_im[v] = True
                self._is_im = True
                self.mylog("IM found", 3)
                if stop_if_true:
                    return
            if nb_wanted_undecided == 0:
                return

    # %% Trivial Manipulation (TM)

    # Use the general methods.

    # %% Ignorant-Coalition Manipulation (ICM)

    # Use the general methods.

    # %% Unison manipulation (UM)

    def _um_main_work_c_(self, c):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [1, 0, 2],
            ...     [1, 2, 0],
            ...     [2, 0, 1],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleMaximin()(profile)
            >>> rule.candidates_um_
            array([0., 1., 1.])
        """
        matrix_duels_temp = preferences_ut_to_matrix_duels_ut(
            self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), :]
        ).astype(float)
        for x in range(self.profile_.n_c):
            matrix_duels_temp[x, x] = np.inf
        scores_temp = np.min(matrix_duels_temp, 1)
        n_m = self.profile_.matrix_duels_ut[c, self.w_]

        w_temp = np.argmax(scores_temp)
        if w_temp == c:  # pragma: no cover
            # TO DO: Investigate whether this case can really happen or not.
            self._reached_uncovered_code()
            self.mylogv("UM: scores_temp =", scores_temp, 3)
            self.mylog("UM: Manipulation easy (c wins without manipulators' votes)", 3)
            self._candidates_um[c] = True
            return
        # Best we can do: improve ``c`` by ``n_m``, do not move the others
        if scores_temp[w_temp] > scores_temp[c] + n_m or (scores_temp[w_temp] >= scores_temp[c] + n_m and w_temp < c):
            self.mylogv("UM: scores_temp =", scores_temp, 3)
            self.mylog("UM: Manipulation impossible (score difference is too high)", 3)
            self._candidates_um[c] = False
            return

        self._vote_strategically_(matrix_duels_temp, scores_temp, c, n_m)
        w_temp = np.argmax(scores_temp)
        self.mylogv("UM: w_temp =", w_temp, 3)
        self._candidates_um[c] = (w_temp == c)

    def sufficient_coalition_size_um_c_(self, c):
        # TODO: put this in cache
        n_s = self.profile_.n_v - self.profile_.matrix_duels_ut[c, self.w_]
        matrix_duels_s = preferences_ut_to_matrix_duels_ut(
            self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), :]
        ).astype(float)
        for x in range(self.profile_.n_c):
            matrix_duels_s[x, x] = np.inf
        scores_s = np.min(matrix_duels_s, 1)
        w_s = np.argmax(scores_s)
        # If w_s == c, we win without any manipulator
        if w_s == c:
            self.mylogv("UM: scores_s =", scores_s, 3)
            self.mylog("UM: Manipulation easy (c wins without manipulators' votes)", 3)
            return 0
        # If we have n_s + 1 manipulators, obviously we can manipulate.
        n_m_sup = n_s + 1
        # Best we can do: improve ``c`` by ``n_m``, do not move the others => we need at least n_m_inf + 1 manipulators.
        n_m_inf = scores_s[w_s] - scores_s[c] - int(c < w_s)
        # Loop invariant: UM always possible n_m_sup manipulators, impossible for n_m_inf manipulators
        while n_m_sup - n_m_inf > 1:
            n_m = (n_m_inf + n_m_sup) // 2
            matrix_duels_temp = matrix_duels_s.copy()
            scores_temp = scores_s.copy()
            self._vote_strategically_(matrix_duels_temp, scores_temp, c, n_m)
            w_temp = np.argmax(scores_temp)
            self.mylogv("UM: w_temp =", w_temp, 3)
            if w_temp == c:
                n_m_sup = n_m
            else:
                n_m_inf = n_m
        return n_m_sup

    # %% Coalition Manipulation (CM)

    def _cm_main_work_c_fast(self, c, optimize_bounds):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 1, 2],
            ...     [1, 2, 0],
            ...     [2, 1, 0],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleMaximin()(profile)
            >>> rule.candidates_cm_
            array([nan,  0.,  0.])
        """
        matrix_duels_temp = preferences_ut_to_matrix_duels_ut(
            self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), :]
        ).astype(float)
        for x in range(self.profile_.n_c):
            matrix_duels_temp[x, x] = np.inf
        scores_temp = np.min(matrix_duels_temp, 1)
        w_temp = np.argmax(scores_temp)
        n_manipulators_used = 0

        # Easy lower bound: for each manipulator, ``c``'s score can at best increase by one, and ``w_temp``'s score
        # cannot decrease.
        self._update_necessary(
            self._necessary_coalition_size_cm, c, scores_temp[w_temp] - scores_temp[c] + (c > w_temp),
            'CM: Update necessary_coalition_size_cm = scores_s[w_s] - scores_s[c] + (c > w_s) =')
        if not optimize_bounds and (self._necessary_coalition_size_cm[c] > self.profile_.matrix_duels_ut[c, self.w_]):
            return True  # is_quick_escape

        # Relatively easy upper bound: if we can UM, we can CM.
        self._update_sufficient(
            self._sufficient_coalition_size_cm, c, self.sufficient_coalition_size_um_c_(c),
            'CM: Update sufficient_coalition_size_cm = sufficient_coalition_size_um =')
        if self._sufficient_coalition_size_cm[c] == self._necessary_coalition_size_cm[c]:
            return False  # Not a quick escape, we did all the job!
        if not optimize_bounds and (self.profile_.matrix_duels_ut[c, self.w_] >= self._sufficient_coalition_size_cm[c]):
            return True  # is_quick_escape

        while w_temp != c:
            self.mylogv("CM: w_temp =", w_temp, 3)
            self.mylogv("CM: c =", c, 3)
            n_manipulators_used += 1
            if n_manipulators_used >= self._sufficient_coalition_size_cm[c]:
                self.mylog("CM: I already know a strategy that works with n_manipulators_used (TM, UM, etc.).")
                self._update_necessary(self._necessary_coalition_size_cm, c, np.ceil(n_manipulators_used * 3 / 5),
                                       'CM: Update necessary_coalition_size_cm =')
                break
            self._vote_strategically_(matrix_duels_temp, scores_temp, c, 1)
            w_temp = np.argmax(scores_temp)
        else:
            self.mylogm("CM: matrix_duels_temp =", matrix_duels_temp, 3)
            self.mylogv("CM: scores_temp =", scores_temp, 3)
            self.mylogv("CM: w_temp =", w_temp, 3)
            self._update_sufficient(self._sufficient_coalition_size_cm, c, n_manipulators_used,
                                    'CM: Update sufficient_coalition_size_cm =')
            # For the 3/5 ratio, see Zuckerman et al.
            self._update_necessary(self._necessary_coalition_size_cm, c, np.ceil(n_manipulators_used * 3 / 5),
                                   'CM: Update necessary_coalition_size_cm =')
        return False

    def _cm_main_work_c_(self, c, optimize_bounds):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleMaximin(cm_option='exact')(profile)
            >>> rule.candidates_cm_
            array([0., 0., 0.])
        """
        is_quick_escape_fast = self._cm_main_work_c_fast(c, optimize_bounds)
        if not self.cm_option == "exact":
            # With 'fast' option, we stop here anyway.
            return is_quick_escape_fast

        # From this point, we have necessarily the 'exact' option (which is, in fact, only an exhaustive exploration
        # with = n_m manipulators).
        is_quick_escape_exact = self._cm_main_work_c_exact_(c, optimize_bounds)
        return is_quick_escape_fast or is_quick_escape_exact


class StackFamily:
    """Family of stacks. Used for the manipulation of Maximin.

    >>> sf = StackFamily()
    >>> print(sf.pop_front())
    None
    >>> sf.push_front(42, 4)
    >>> sf.push_front(51, 0)
    >>> sf.push_front(42, 2)
    >>> sf.push_front(42, 3)
    >>> sf
    {42: [4, 2, 3], 51: [0]}
    >>> sf.pop_front()
    3
    >>> sf
    {42: [4, 2], 51: [0]}

    A StackFamily contains a dictionary whose keys are numbers and values are lists.
    """

    def __init__(self):
        self.data = dict()

    def __repr__(self):
        return '{' + ', '.join(['%s: %s' % (k, self.data[k]) for k in sorted(self.data.keys())]) + '}'

    def push_front(self, score, candidate):
        """Push a value to the appropriate stack (given by 'score').
        """
        try:
            self.data[score].append(candidate)
        except KeyError:
            self.data[score] = [candidate]

    def pop_front(self):
        """Pop a value: choose the stack with minimum score, and pop the most recent item in this stack.

        When the object is empty, returns None.
        """
        if len(self.data) == 0:
            return None
        lowest_score = min(self.data.keys())
        c = self.data[lowest_score].pop()
        if len(self.data[lowest_score]) == 0:
            del self.data[lowest_score]
        return c



from svvamp import RuleIRV, Profile


def test_fast_algo_setter():
    """
        >>> rule = RuleIRV()
        >>> rule.fast_algo = 'foo'
        Traceback (most recent call last):
        ValueError: Unknown value for option fast_algo: foo
    """
    pass


def test_log_um_():
    """
        >>> rule = RuleIRV(um_option='exact')
        >>> rule.log_um_
        'um_option = exact'
    """
    pass


def test_log_cm_():
    """
        >>> rule = RuleIRV(cm_option='exact')
        >>> rule.log_cm_
        'cm_option = exact'
    """
    pass


def test_v_might_be_pivotal():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleIRV()(profile)
        >>> rule.v_might_be_pivotal_
        array([ True,  True,  True,  True,  True])
    """
    pass


def test_v_might_im_for_c():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleIRV()(profile)
        >>> rule.v_might_im_for_c_
        array([[ True,  True,  True],
               [ True,  True,  True],
               [ True,  True,  True],
               [ True,  True,  True],
               [ True,  True,  True]])
    """
    pass


def test_is_not_iia():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleIRV()(profile)
        >>> rule.is_not_iia_
        False
    """
    pass


def test_is_icm_c():
    """
    Examples
    --------
        >>> profile = Profile(preferences_rk=[
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [1, 2, 0],
        ...     [2, 1, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleIRV()(profile)
        >>> rule.is_icm_c_(0)
        False
        >>> rule.is_icm_c_with_bounds_(0)
        (False, 5.0, 5.0)
    """
    pass



# -*- coding: utf-8 -*-
"""
Created on jul. 21, 2021, 15:26
Copyright François Durand 2014-2021
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from svvamp.utils.misc import indent
from svvamp.rules.rule_irv import RuleIRV


class StudyRuleCriteria:
    """A set of criteria to study for the simulator about one or several voting rules.

    Parameters
    ----------
    manipulation_criteria : List of str
        Manipulation criteria that need to be studied. If None (default), then all criteria related to
        collective manipulation are taken: ['is_tm_', 'is_um_', 'is_icm_', 'is_cm_']. If [], then no criterion is taken.
        Note that by default, 'is_iia_' and 'is_im_' are NOT included.
    manipulation_criteria_c : List of str
        This is where you put criteria such as 'candidates_cm_', 'candidates_tm_', etc.
    manipulation_only : bool
        If True, then result_criteria and utility criteria are ignored, i.e. set to [].
    result_criteria : List of str
        Result criteria that need to be studied. If None(default), then all criteria are taken:
        ['w_is_weak_condorcet_winner_', 'w_missed_weak_condorcet_winner_', ...].
    utility_criteria : List
        List of tuples  (criterion, func, name), where:
        criterion: the numerical data to be processed.
        func: an aggregating function.
        name: the name chosen for this pair criterion / aggregation function.
        Example: ('total_utility_w_', np.mean, 'total_u_mean').
        Will compute the mean of total_utility_w over the populations and the result will be called 'total_u_mean'.
        If utility_criteria is None (default), then use:
        [('total_utility_w_', np.max, 'total_u_max'),
         ('total_utility_w_', np.min, 'total_u_min'),
         ('total_utility_w_', np.mean, 'total_u_mean'),
         ('total_utility_w_', np.std, 'total_u_std')].
    numerical_criteria : List of str
        Other numerical criteria that need to be studied. These criteria must correspond to attributes in the rule
        that output a pair (inf, sup), for example `nb_candidates_cm_` or `worst_relative_welfare_with_cm_`.
    """

    def __init__(self, manipulation_criteria=None, manipulation_criteria_c=None, manipulation_only=False,
                 result_criteria=None, utility_criteria=None, numerical_criteria=None):
        # Manipulation criteria
        if manipulation_criteria is None:
            manipulation_criteria = [
                'is_tm_',
                'is_um_',
                'is_icm_',
                'is_cm_',
                'is_tm_or_um_',
                'elects_condorcet_winner_rk_even_with_cm_',
            ]
        self.manipulation_criteria = manipulation_criteria
        if manipulation_criteria_c is None:
            manipulation_criteria_c = []
        self.manipulation_criteria_c = manipulation_criteria_c
        # Result criteria
        if manipulation_only:
            self.result_criteria = []
        elif result_criteria is None:
            self.result_criteria = [
                'w_is_condorcet_admissible_',
                'w_is_not_condorcet_admissible_',
                'w_missed_condorcet_admissible_',
                'w_is_weak_condorcet_winner_',
                'w_is_not_weak_condorcet_winner_',
                'w_missed_weak_condorcet_winner_',
                'w_is_condorcet_winner_rk_ctb_',
                'w_is_not_condorcet_winner_rk_ctb_',
                'w_missed_condorcet_winner_rk_ctb_',
                'w_is_condorcet_winner_rk_',
                'w_is_not_condorcet_winner_rk_',
                'w_missed_condorcet_winner_rk_',
                'w_is_condorcet_winner_ut_rel_ctb_',
                'w_is_not_condorcet_winner_ut_rel_ctb_',
                'w_missed_condorcet_winner_ut_rel_ctb_',
                'w_is_condorcet_winner_ut_rel_',
                'w_is_not_condorcet_winner_ut_rel_',
                'w_missed_condorcet_winner_ut_rel_',
                'w_is_condorcet_winner_ut_abs_ctb_',
                'w_is_not_condorcet_winner_ut_abs_ctb_',
                'w_missed_condorcet_winner_ut_abs_ctb_',
                'w_is_condorcet_winner_ut_abs_',
                'w_is_not_condorcet_winner_ut_abs_',
                'w_missed_condorcet_winner_ut_abs_',
                'w_is_resistant_condorcet_winner_',
                'w_is_not_resistant_condorcet_winner_',
                'w_missed_resistant_condorcet_winner_'
            ]
        else:
            self.result_criteria = result_criteria
        # Utility criteria
        if manipulation_only:
            self.utility_criteria = []
        elif utility_criteria is None:
            self.utility_criteria = [
                ('mean_utility_w_', np.max, 'mean_u_max'),
                ('mean_utility_w_', np.min, 'mean_u_min'),
                ('mean_utility_w_', np.mean, 'mean_u_mean'),
                ('mean_utility_w_', np.std, 'mean_u_std'),
                ('relative_social_welfare_w_', np.mean, 'relative_social_welfare_mean'),
            ]
        else:
            self.utility_criteria = utility_criteria
        # Numerical criteria
        if numerical_criteria is None:
            self.numerical_criteria = [
                'nb_candidates_cm_',
                'worst_relative_welfare_with_cm_',
            ]
        else:
            self.numerical_criteria = numerical_criteria

    def __str__(self):
        """
            >>> study_rule_criteria = StudyRuleCriteria(
            ...     manipulation_criteria=['is_cm_'],
            ...     manipulation_criteria_c=['candidates_cm_'],
            ...     result_criteria=['w_is_condorcet_admissible_', 'w_is_weak_condorcet_winner_'],
            ...     utility_criteria=[('mean_utility_w_', np.max, 'mean_u_max'),
            ...                       ('mean_utility_w_', np.min, 'mean_u_min')],
            ...     numerical_criteria=['nb_candidates_cm_', 'worst_relative_welfare_with_cm_'],
            ... )
            >>> print(study_rule_criteria)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
            StudyRuleCriteria with:
                manipulation_criteria:
                    is_cm_
                manipulation_criteria_c:
                    candidates_cm_
                result_criteria:
                    w_is_condorcet_admissible_
                    w_is_weak_condorcet_winner_
                utility_criteria:
                    ('mean_utility_w_', <function amax at ...>, 'mean_u_max')
                    ('mean_utility_w_', <function amin at ...>, 'mean_u_min')
                numerical_criteria:
                    nb_candidates_cm_
                    worst_relative_welfare_with_cm_

            >>> study_rule_criteria = StudyRuleCriteria(
            ...     manipulation_criteria=[],
            ...     result_criteria=[],
            ...     utility_criteria=[],
            ...     numerical_criteria=[],
            ... )
            >>> print(study_rule_criteria)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
            StudyRuleCriteria with:
                manipulation_criteria: None
                manipulation_criteria_c: None
                result_criteria: None
                utility_criteria: None
                numerical_criteria: None
        """
        # noinspection PyListCreation
        lines = []
        for name in ['manipulation_criteria', 'manipulation_criteria_c', 'result_criteria',
                     'utility_criteria', 'numerical_criteria']:
            criteria = getattr(self, name)
            if criteria:
                lines.append(f'{name}:')
                lines.extend(['    ' + str(criterion) for criterion in criteria])
            else:
                lines.append(f'{name}: None')
        return 'StudyRuleCriteria with:\n' + indent('\n'.join(lines))

    def remove(self, criterion):
        """Remove a criterion

        Remove all occurrences of a criterion (generally zero or one) from this `StudyRuleCriteria` object.

        Parameters
        ----------
        criterion : str

        Returns
        -------
        StudyRuleCriteria
            The object itself.

        Examples
        --------
            >>> study_rule_criteria = StudyRuleCriteria(
            ...     manipulation_criteria=['is_cm_', 'is_tm_'],
            ...     result_criteria=['w_is_condorcet_admissible_', 'w_is_weak_condorcet_winner_'],
            ...     utility_criteria=[('mean_utility_w_', np.max, 'mean_u_max'),
            ...                       ('mean_utility_w_', np.min, 'mean_u_min')],
            ...     numerical_criteria=['nb_candidates_cm_', 'worst_relative_welfare_with_cm_'],
            ... )
            >>> study_rule_criteria.remove('is_cm_')
            >>> study_rule_criteria.remove('w_is_weak_condorcet_winner_')
            >>> study_rule_criteria.remove('mean_u_max')
            >>> study_rule_criteria.remove('worst_relative_welfare_with_cm_')
            >>> print(study_rule_criteria)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
            StudyRuleCriteria with:
                manipulation_criteria:
                    is_tm_
                manipulation_criteria_c: None
                result_criteria:
                    w_is_condorcet_admissible_
                utility_criteria:
                    ('mean_utility_w_', <function amin at ...>, 'mean_u_min')
                numerical_criteria:
                    nb_candidates_cm_
        """
        self.manipulation_criteria = [c for c in self.manipulation_criteria if c != criterion]
        self.manipulation_criteria_c = [c for c in self.manipulation_criteria_c if c != criterion]
        self.result_criteria = [c for c in self.result_criteria if c != criterion]
        self.utility_criteria = [c for c in self.utility_criteria if c[2] != criterion]
        self.numerical_criteria = [c for c in self.numerical_criteria if c != criterion]

    def append_manipulation_criterion(self, criterion):
        """Add a manipulation criterion

        Parameters
        ----------
        criterion : str

        Examples
        --------
            >>> study_rule_criteria = StudyRuleCriteria(
            ...     manipulation_criteria=[],
            ...     result_criteria=[],
            ...     utility_criteria=[],
            ...     numerical_criteria=[],
            ... )
            >>> study_rule_criteria.append_manipulation_criterion('is_im_')
            >>> study_rule_criteria.append_manipulation_criterion('is_iia_')
            >>> print(study_rule_criteria)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
            StudyRuleCriteria with:
                manipulation_criteria:
                    is_im_
                    is_iia_
                manipulation_criteria_c: None
                result_criteria: None
                utility_criteria: None
                numerical_criteria: None
        """
        if criterion not in self.manipulation_criteria:
            self.manipulation_criteria.append(criterion)

    def append_manipulation_criterion_c(self, criterion):
        """Add a manipulation criterion (for candidates)

        Parameters
        ----------
        criterion : str

        Examples
        --------
            >>> study_rule_criteria = StudyRuleCriteria(
            ...     manipulation_criteria=[],
            ...     manipulation_criteria_c=[],
            ...     result_criteria=[],
            ...     utility_criteria=[],
            ...     numerical_criteria=[],
            ... )
            >>> study_rule_criteria.append_manipulation_criterion_c('candidates_cm_')
            >>> study_rule_criteria.append_manipulation_criterion_c('candidates_tm_')
            >>> print(study_rule_criteria)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
            StudyRuleCriteria with:
                manipulation_criteria: None
                manipulation_criteria_c:
                    candidates_cm_
                    candidates_tm_
                result_criteria: None
                utility_criteria: None
                numerical_criteria: None
        """
        if criterion not in self.manipulation_criteria_c:
            self.manipulation_criteria_c.append(criterion)

    def append_result_criterion(self, criterion):
        """Add a result criterion

        Parameters
        ----------
        criterion : str

        Examples
        --------
            >>> study_rule_criteria = StudyRuleCriteria(
            ...     manipulation_criteria=[],
            ...     result_criteria=[],
            ...     utility_criteria=[],
            ...     numerical_criteria=[],
            ... )
            >>> study_rule_criteria.append_result_criterion('w_is_condorcet_admissible_')
            >>> print(study_rule_criteria)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
            StudyRuleCriteria with:
                manipulation_criteria: None
                manipulation_criteria_c: None
                result_criteria:
                    w_is_condorcet_admissible_
                utility_criteria: None
                numerical_criteria: None
        """
        if criterion not in self.result_criteria:
            self.result_criteria.append(criterion)

    def append_utility_criterion(self, criterion, func, name):
        """Add a utility criterion

        Parameters
        ----------
        criterion: str
            The numerical data to be processed.
        func: callable
            An aggregating function.
        name: str
            The name chosen for this pair criterion / aggregation function.

        Examples
        --------
            >>> study_rule_criteria = StudyRuleCriteria(
            ...     manipulation_criteria=[],
            ...     result_criteria=[],
            ...     utility_criteria=[],
            ...     numerical_criteria=[],
            ... )
            >>> study_rule_criteria.append_utility_criterion('total_utility_w_', np.sum, 'total_u_sum')
            >>> print(study_rule_criteria)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
            StudyRuleCriteria with:
                manipulation_criteria: None
                manipulation_criteria_c: None
                result_criteria: None
                utility_criteria:
                    ('total_utility_w_', <function sum at ...>, 'total_u_sum')
                numerical_criteria: None

        In the above example, the criterion will be used to compute the sum of `total_utility_w_` over all the profiles,
        and the result will be called 'total_u_sum'.
        """
        if name not in [n for (c, f, n) in self.utility_criteria]:
            self.utility_criteria.append((criterion, func, name))

    def append_numerical_criterion(self, criterion):
        """Add a numerical criterion

        Parameters
        ----------
        criterion : str

        Examples
        --------
            >>> study_rule_criteria = StudyRuleCriteria(
            ...     manipulation_criteria=[],
            ...     result_criteria=[],
            ...     utility_criteria=[],
            ...     numerical_criteria=[],
            ... )
            >>> study_rule_criteria.append_numerical_criterion('nb_candidates_cm_')
            >>> print(study_rule_criteria)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
            StudyRuleCriteria with:
                manipulation_criteria: None
                manipulation_criteria_c: None
                result_criteria: None
                utility_criteria: None
                numerical_criteria:
                    nb_candidates_cm_
        """
        if criterion not in self.numerical_criteria:
            self.numerical_criteria.append(criterion)

    def check_sanity(self, rule_class):
        """Check sanity of the object

        Preform some basic checks. It is recommended to use this method before launching big simulations.

        Parameters
        ----------
        rule_class : class
            Subclass of :class:`Rule`.

        Examples
        --------
        Detect an illegal manipulation criterion:

            >>> study_rule_criteria = StudyRuleCriteria(manipulation_criteria=['unexpected_criterion'])
            >>> study_rule_criteria.check_sanity(rule_class=RuleIRV)
            Traceback (most recent call last):
            ValueError: Criterion 'unexpected_criterion' is unknown for RuleIRV.

        Detect an illegal manipulation criterion (for candidates):

            >>> study_rule_criteria = StudyRuleCriteria(manipulation_criteria_c=['unexpected_criterion'])
            >>> study_rule_criteria.check_sanity(rule_class=RuleIRV)
            Traceback (most recent call last):
            ValueError: Criterion 'unexpected_criterion' is unknown for RuleIRV.

        Detect an illegal result criterion:

            >>> study_rule_criteria = StudyRuleCriteria(
            ...     manipulation_criteria=[],
            ...     result_criteria=['unexpected_criterion'])
            >>> study_rule_criteria.check_sanity(rule_class=RuleIRV)
            Traceback (most recent call last):
            ValueError: Criterion 'unexpected_criterion' is unknown for RuleIRV.

        Detect an illegal utility criterion:

            >>> study_rule_criteria = StudyRuleCriteria(
            ...     manipulation_criteria=[],
            ...     utility_criteria=[('unexpected_criterion', np.sum, 'the_name')])
            >>> study_rule_criteria.check_sanity(rule_class=RuleIRV)
            Traceback (most recent call last):
            ValueError: Criterion 'unexpected_criterion' is unknown for RuleIRV.

        Detect when a aggregation function is not provided:

            >>> study_rule_criteria = StudyRuleCriteria(
            ...     manipulation_criteria=[],
            ...     utility_criteria=[('mean_utility_w_', 'this_is_not_callable', 'the_name')])
            >>> study_rule_criteria.check_sanity(rule_class=RuleIRV)
            Traceback (most recent call last):
            TypeError: Expected: callable. Got: 'this_is_not_callable'.

        Detect an illegal numerical criterion:

            >>> study_rule_criteria = StudyRuleCriteria(
            ...     manipulation_criteria=[],
            ...     numerical_criteria=['unexpected_criterion'])
            >>> study_rule_criteria.check_sanity(rule_class=RuleIRV)
            Traceback (most recent call last):
            ValueError: Criterion 'unexpected_criterion' is unknown for RuleIRV.
        """
        # Check if manipulation criteria are legal.
        for criterion in self.manipulation_criteria:
            if criterion not in dir(rule_class):
                raise ValueError(f"Criterion '{criterion}' is unknown for {rule_class.__name__}.")
        # Check if manipulation criteria (candidates) are legal.
        for criterion in self.manipulation_criteria_c:
            if criterion not in dir(rule_class):
                raise ValueError(f"Criterion '{criterion}' is unknown for {rule_class.__name__}.")
        # Check if result criteria are legal.
        for criterion in self.result_criteria:
            if criterion not in dir(rule_class):
                raise ValueError(f"Criterion '{criterion}' is unknown for {rule_class.__name__}.")
        # Check if utility criteria are legal.
        for (criterion, func, name) in self.utility_criteria:
            if criterion not in dir(rule_class):
                raise ValueError(f"Criterion '{criterion}' is unknown for {rule_class.__name__}.")
            if not hasattr(func, '__call__'):
                raise TypeError(f"Expected: callable. Got: {repr(func)}.")
        # Check if numerical criteria are legal.
        for criterion in self.numerical_criteria:
            if criterion not in dir(rule_class):
                raise ValueError(f"Criterion '{criterion}' is unknown for {rule_class.__name__}.")
